#include <ctime>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <map>
#include <set>
#include <omp.h>

/*
#include "GLOBAL.h"
#include "Utils.h"
#include "CtgsReads.h"
#include "BaseStr.h"
#include "AccTester.h"
#include "MetaCluster.h"
#include "USet.h"
*/

//#include "GenomeDB_KmerDistri.h"
#include "Methods.h"
#include "MCPara.h"
#include "DBEntropy.h"
using namespace std;
/////////////////////////////////////////////////////////////////////
static const unsigned UNASSIGNED  = ~(0U);
const double ScoreThresh = -1.0;
//parameters
int ReadLen = 80;
int MetaKmerLen = 5;
int ClusterSize = 0;
int MaxSpecies = 200;
int MinSpecies = 2;
int StrLen = 50;
int StrLenLowCover = 25;
double MC3_Thresh = 0.94;
int OrphanLen = 400;

int Num_Thread = 0;
int CtgLenThresh = 500;
int AlignThresh = 76;
/////////////////////////////////////////////////////////////////////
time_t rawtime;
struct tm* timeinfo;

KmerNode** KmerMap;
ContigsClass Ctgs;
ReadsClass Reads;
KmerNodeAloc NodePool;
USet uset;
//GenomesClass GenoDB;
DBEntropy GenoDB;
vector<TaxoInfoClass> TaxoInfo;
vector<ClustTaxoInfoClass> TaxoOfClust;

AccTester acc_tester;

////////////////////////////////////////////////////////////////////////////
void usage()
{
	cerr << "usage:\t MetaCluster_HB contig.fa read.fa GenomeFile.txt [options]" << endl;
	cerr << endl;
	exit(-1);
}

void printtime(string str = "")
{
	cerr << str << endl;
	time(&rawtime);
	timeinfo = localtime(&rawtime);
	cerr<<asctime(timeinfo);
}

void init()
{
	///////////////////////////////////////////////////////
	{
		unsigned long long u64Exp = ((((1ULL << 50)%HASHSIZE)<<14))%HASHSIZE;
		unsigned long long tExp = 1;
		for(int i=0;i<10;++i)
		{
			U64HASH[i] = tExp;
			tExp *= u64Exp;
			tExp %= HASHSIZE;
		}
	}
	KmerMap = new KmerNode*[HASHSIZE];
	for(unsigned i=0;i<HASHSIZE;++i)
		KmerMap[i] = NULL;
	///////////////////////////////////////////////////////
}

void input(int argc, char* argv[])
{
	if(argc<4)
		usage();
	//Important! readlength should belong to (50,128]. Default value = 75.
	AddParameter("ReadLen", &ReadLen, INTEGER);
	//We use q-mer distribution in Phase 3. Here you can set qmer = 4 or 5. Default value = 4.
	AddParameter("qmer", &MetaKmerLen, INTEGER);
	//You can fix the #. of species. Default value = 0. 
	//In this case, the program will predict the #. of species.
	AddParameter("Species", &ClusterSize, INTEGER);
	//We use binary search to predict #. of species, this is max value for the first iteration.
	//Set this number properly will improve the performance and speed up the program. Default = 600.
	AddParameter("MaxSpecies", &MaxSpecies, INTEGER);
	//We use binary search to predict #. of species, this is min value for the first iteration. 
	//Set this number properly will improve the performance and speed up the program. Default = 2.
	AddParameter("MinSpecies", &MinSpecies, INTEGER);
	//wmer in phase 1, default = 50;
	AddParameter("wmer", &StrLen, INTEGER);
	//wmer for low coverage read , default = 25; if a smaller value is needed, mapinto() should be modifed to mappint more 16-mers.
	AddParameter("lowwmer", &StrLenLowCover, INTEGER);
	//Threshold for further-merge in MetaCluster 3.0. Default valule = 0.9. If you set this value larger, program will conceptually predict more species. 
	AddParameter("MetaC3Threshold", &MC3_Thresh, FLOAT);
	//groups of size smaller than Orphan16merThresh will be considered as orphans
	AddParameter("Orphan", &OrphanLen, INTEGER);
	//////////////////////////////////////////////////////////////////////////
	//set minimum length of contigs.
	AddParameter("CtgLenThresh", &CtgLenThresh, INTEGER);
	AlignThresh = ReadLen*0.95;
	AddParameter("AlignThresh", &AlignThresh, INTEGER);

	ProcessParameters(argc,argv);
	/////////////////////////////////////////////////////////////////
	cerr << dec << "ReadLen:\t " << ReadLen << endl;
	cerr << "CtgLenThresh:\t " << CtgLenThresh << endl;
	cerr << "AlignThresh:\t " << AlignThresh << endl;
	cerr << "MC3_Thresh:\t" << MC3_Thresh << endl;
	/////////////////////////////////////////////////////////////////
	if(INTEST)printtime("before loading contigs. ");
	Ctgs.init(argv[1], CtgLenThresh,KmerMap, NodePool);
	if(INTEST)
	{
		cerr << "Contigs: \t" << Ctgs.CtgNum << endl;
		/*
		map<int,int>MId;
		map<int,int>MOc;
		omp_lock_t tlock;
		omp_init_lock(&tlock);
#pragma omp parallel for
		for(unsigned i=0;i<HASHSIZE;++i)
		{
			for(KmerNode* p = KmerMap[i];p!=NULL;p=p->next)
			{
				set<int>ids;
				for(int j=0;j<p->VSize;++j)
					ids.insert(p->myvector[j].id);
				omp_set_lock(&tlock);
				++MId[ids.size()];
				++MOc[p->VSize];
				omp_unset_lock(&tlock);
			}
		}
		cout << "Id Occurence: " << endl;
		for(map<int,int>::const_iterator itr=MId.begin();itr!=MId.end();++itr)
			cout << itr->first << '\t' << itr->second << endl;
		cout << "Kmer Occurence: " << endl;
		for(map<int,int>::const_iterator itr=MOc.begin();itr!=MOc.end();++itr)
			cout << itr->first << '\t' << itr->second << endl;
		*/
		printtime("Contigs loaded. ");
		///////////////////////////////////////////////////////////
	}
	NodePool.fixCtgNum();
	if(INTEST)cerr<<"Ctg Num fixed." << endl;
	////////////////////////////////
//	GenoDB.init(argv[3], KmerMap, NodePool);
	GenoDB.init(argv[3]);
	system("ps ux");
	if(INTEST)cerr<<"Genome DB loaded: "<< Ctgs.CtgNum << endl;
	TaxoInfo.resize(Ctgs.CtgNum);
	if(INTEST)cerr<<"Before Annotating contigs:" << endl;
#pragma omp parallel for
	for(int i=0;i<Ctgs.CtgNum;++i)
	{
		unsigned taxid=UNASSIGNED;double score;
		unsigned simulateId;double simulateScore;
		vector<double>entropy;
		GenoDB.calTaxoForCtg(Ctgs.contigs[i]->str, taxid, score,entropy,simulateId,simulateScore);
		if(score < 0)taxid = UNASSIGNED;
		TaxoInfo[i].set(taxid,score);
/*		vector<unsigned> taxid; vector<double> score;
		GenoDB.calTaxoForCtg(Ctgs.contigs[i]->str, taxid, score);
		for(int j=0;j<7;++j)cout << score[j] << '\t';
		*/
		if(score>=0)
		{
			cout << i << '\t' << taxid << '\t' << score <<'\t'<<simulateScore << '\t' << GenoDB.GenomeLength[taxid] << '\t';
			cout << GenoDB.GenomeLength[simulateId] << '\t';
			for(int j=0;j<7;++j)cout << GenoDB.Taxo[taxid][j] << '\t';
			for(int j=0;j<7;++j)cout << entropy[j] << '\t';
			cout << Ctgs.contigs[i]->length <<'\t'<< Ctgs.info[i] << endl;
		}
	//	cout << taxid << '\t' << taxLevel << '\t' << score << '\t' << Ctgs.info[i] << endl;
	}
//	GenoDB.getStatis(NodePool);
//	exit(-1);
	/////////////////////////////////////////////////////////////////
	KmerDistriPara Para5mer(5);
	int** ctgSpear = Ctgs.getSpear(Para5mer);
	for(int i=0;i<Ctgs.CtgNum;++i)
	{
		ctgSpear[i] = NULL;
	}
	////////////////////////////////
	printtime("initializing reads. ");
	Reads.init(argv[2], ReadLen, AlignThresh, KmerMap, NodePool, Ctgs, acc_tester);
	{
		printtime("initializing uset: ");
		int CtgNum = Ctgs.CtgNum;
		unsigned* ctglen = new unsigned[CtgNum];
		for(int i=0;i<CtgNum;++i)
			ctglen[i] = Ctgs.contigs[i]->length;
		uset.init(Ctgs.CtgNum + Reads.ReadNum, CtgNum, ctglen);
		delete[]ctglen;
	}

	if(INTEST)
	{
		printtime("Reads loaded. ");
		cerr << dec << "Total Reads:\t"<< Reads.TotalNum << endl;
		cerr << "Unmaped Reads:\t" << Reads.ReadNum << endl;
		cerr << "AlignThresh:\t" << Reads.AlignThresh << endl;
		/*
		map<int,int>MId;
		map<int,int>MOc;
		omp_lock_t tlock;
		omp_init_lock(&tlock);
#pragma omp parallel for
		for(unsigned i=0;i<HASHSIZE;++i)
		{
			for(KmerNode* p = KmerMap[i];p!=NULL;p=p->next)
			{
				set<int>ids;
				for(int j=0;j<p->VSize;++j)
					ids.insert(p->myvector[j].id);
				omp_set_lock(&tlock);
				++MId[ids.size()];
				++MOc[p->VSize];
				omp_unset_lock(&tlock);
			}
		}
		cout << "Id Occurence: " << endl;
		for(map<int,int>::const_iterator itr=MId.begin();itr!=MId.end();++itr)
			cout << itr->first << '\t' << itr->second << endl;
		cout << "Kmer Occurence: " << endl;
		for(map<int,int>::const_iterator itr=MOc.begin();itr!=MOc.end();++itr)
			cout << itr->first << '\t' << itr->second << endl;
		///////////////////////////////////////////////////////////////
		cout << "Score Distribution." << endl;
		map<int,int>MS;
		for(int i=0;i<Reads.TotalNum;++i)
			++MS[Reads.Score[i]];
		for(map<int,int>::const_iterator itr=MS.begin();itr!=MS.end();++itr)
			cout << itr->first << '\t' << itr->second << endl;
		///////////////////////////////////////////////////////////////
		cout << "Coverage Distribution:" << endl;
		map<int,int>Cov;
		for(int i=0;i<Reads.TotalNum;++i)
			if(Reads.MatchId[i]>=0)
				++Cov[Reads.MatchId[i]];
		map<int,int>CDis;
		for(map<int,int>::const_iterator itr=Cov.begin();itr!=Cov.end();++itr)
			++CDis[(itr->second)*ReadLen/(1*(Ctgs.contigs[itr->first]->length))];
		for(map<int,int>::const_iterator itr=CDis.begin();itr!=CDis.end();++itr)
			cout << itr->first*1 << '\t' << itr->second << endl;
		*/
		///////////////////////////////////////////////////////////////////
		printtime("initializing acc_tester:\t");
		acc_tester.init(Reads.TotalNum, Reads.MatchId, Reads.ReadNum,Reads.NewIdToOldId);
	/*	acc_tester.calAcc(uset);*/
	}
	delete[] KmerMap;
	KmerMap = NULL;
//	NodePool.shrinkSize(NULL,2);
	//////////////////////////////////////////////////////////
}

void anaCluster(MCPara& mcpara, MetaCluster& metacluster,vector<unsigned>& ReadNumInCtg)
{
	int VCtgNum = metacluster.Size;
	int ClustNum = 0;
	for(int i=0;i<VCtgNum;++i)
		ClustNum = max(metacluster.best[i],ClustNum);
	++ClustNum;
	vector<int>toClustId(Ctgs.CtgNum,0);
	for(int i=0;i<Ctgs.CtgNum;++i)
		toClustId[i] = metacluster.best[mcpara.getNewId(uset.find(i))];
	cout << "Cluster Id of Ctgs:\t"<< endl; 
	for(int i=0;i<Ctgs.CtgNum;++i)
		cout << toClustId[i] << endl;
	vector<vector<int> >CtgIdInCluster(ClustNum);
	for(int i=0;i<Ctgs.CtgNum;++i)
		CtgIdInCluster[toClustId[i]].push_back(i);
	TaxoOfClust.resize(ClustNum);
	///////////////////////////////////////////////////////////////////////
	//Taxo:
/*	for(vector<vector<int> >::const_iterator itr1= CtgIdInCluster.begin();itr1!=CtgIdInCluster.end();++itr1)
	{
		for(int level = 6;level >= 0;--level)
		{
			map<int,long long>comp;
			for(vector<int>::const_iterator itr2 = itr1->begin();itr2!=itr1->end();++itr2)
			{
				if(TaxoInfo[*itr2].taxid != UNASSIGNED)
					comp[GenoDB.Taxo[TaxoInfo[*itr2].taxid][level]] += (Ctgs.contigs[*itr2])->length;//ReadNumInCtg[*itr2];
				else
					comp[0] += (Ctgs.contigs[*itr2])->length;//ReadNumInCtg[*itr2];
			}
			double score = getScoreEntropy(comp);
			if(score >= ScoreThresh)
			{
				break;
			}
		}
	}*/
	///////////////////////////////////////////////////////////////////////

	string TName[] = {"kindom","phylum","class","order","family","genus","species"};
	for(int level=0;level<7;++level)
	{
		cout << "Num of clusters: \t" << CtgIdInCluster.size() << endl;
		cout << "taxo for clusters in "<<TName[level]<<" level." << endl;
		int idx = 0;
		for(vector<vector<int> >::const_iterator itr1= CtgIdInCluster.begin();itr1!=CtgIdInCluster.end();++itr1)
		{
			map<int,long long>comp;
			for(vector<int>::const_iterator itr2 = itr1->begin();itr2!=itr1->end();++itr2)
			{
				if(TaxoInfo[*itr2].taxid != UNASSIGNED)
					comp[GenoDB.Taxo[TaxoInfo[*itr2].taxid][level]] += (Ctgs.contigs[*itr2])->length;//ReadNumInCtg[*itr2];
				//	comp[GenoDB.Taxo[TaxoInfo[*itr2].taxid][4]]+=ReadNumInCtg[*itr2];
				else
					comp[0] += (Ctgs.contigs[*itr2])->length;//ReadNumInCtg[*itr2];
				//	comp[0]+=ReadNumInCtg[*itr2];
			}
			//	++comp[GenoDB.Taxo[TaxoInfo[*itr2].taxid][4]];
	
			int maxid=0,maxTaxoid=0;
			long long readsInCluster = 0;
			for(map<int,long long>::const_iterator itr2=comp.begin();itr2!=comp.end();++itr2)
			{
				if(itr2->second > maxid)
				{
					maxid = itr2->second;
					maxTaxoid = itr2->first;
				}
				readsInCluster += itr2->second;
			}
			double sumscore = 0;
			for(vector<int>::const_iterator itr2 = itr1->begin();itr2!=itr1->end();++itr2)
			{
				if(TaxoInfo[*itr2].taxid != UNASSIGNED)
				{
					if(GenoDB.Taxo[TaxoInfo[*itr2].taxid][level] == maxTaxoid)
						sumscore += TaxoInfo[*itr2].score;
				}
				else
				{
					if(0 == maxTaxoid)
						sumscore += TaxoInfo[*itr2].score;
				}
			}

			TaxoOfClust[idx].taxo[level].set(maxTaxoid,maxid/(double)readsInCluster);
			TaxoOfClust[idx].alignscore[level]=sumscore;
			if(itr1->size()>0)
				cout << idx << '\t' << maxTaxoid << '\t' << maxid << '\t' << readsInCluster << '\t' <<sumscore << '\t' << itr1->size()<<'\t'<<getScoreEntropy(comp) <<'\t'<<getScoreMax(comp)<<'\t'<<getScoreMax2(comp) << endl;;
			++idx;
		}
	}

	if(INTEST)
	{
		cout << "detail taxo for clusters in species level." << endl;
		int idx = 0;
		for(vector<vector<int> >::const_iterator itr1= CtgIdInCluster.begin();itr1!=CtgIdInCluster.end();++itr1)
		{
			map<int,int>comp;
			long long readsInCluster = 0;
			for(vector<int>::const_iterator itr2 = itr1->begin();itr2!=itr1->end();++itr2)
			{
				if(TaxoInfo[*itr2].taxid != UNASSIGNED)
					comp[GenoDB.Taxo[TaxoInfo[*itr2].taxid][6]] += (Ctgs.contigs[*itr2])->length;//ReadNumInCtg[*itr2];
				//	comp[GenoDB.Taxo[TaxoInfo[*itr2].taxid][6]]+=ReadNumInCtg[*itr2];
				else
					comp[0] += (Ctgs.contigs[*itr2])->length;//ReadNumInCtg[*itr2];
				//	comp[0]+=ReadNumInCtg[*itr2];
			}
			//	++comp[GenoDB.Taxo[TaxoInfo[*itr2].taxid][6]];
			if(itr1->size()<=0)continue;
			cout << idx << '\t' <<readsInCluster << '\t' << itr1->size() << '\t';
			++idx;
			for(map<int,int>::const_iterator itr2=comp.begin();itr2!=comp.end();++itr2)
				cout << itr2->first << '\t' << itr2->second << '\t';
			cout << endl;
		}
		cout << endl << endl;
	}
}

int main(int argc, char* argv[])
{
	if(argc < 4)
		usage();
	init();
	input(argc,argv);

	MergeAsStep1(Ctgs,Reads,NodePool,uset,acc_tester,AlignThresh);

	if(INTEST)acc_tester.calAcc(uset);

	MCPara mcpara(MetaKmerLen, Ctgs, uset, acc_tester);
	MetaCluster metacluster(mcpara.KmerLen, mcpara.Size, mcpara.ReverKmerDistri, ClusterSize, MaxSpecies, MinSpecies, mcpara.GenoNum , mcpara.Component);
	ClusterSize ? (metacluster.muiltkmeans(10,ClusterSize)) : (metacluster.iterMeta(10,MC3_Thresh));
	anaCluster(mcpara, metacluster, Reads.ReadNumInCtg);
	/* ToDo: clear memory
	 * because outputResult needs O(n)memory, where n is # reads.
	 */
	outputResult(Reads.MatchId, Reads.TotalNum, uset,metacluster.best, mcpara,argv[2], TaxoOfClust);

	if(INTEST)
	{
		printtime("Clustering finished. ");
		acc_tester.getPreSen4Other(metacluster.getComp(),metacluster.Size,metacluster.best);
	}
	return 0;
}
